#!/usr/bin/env python

import os
import sys
from math import *
from copy import copy

from itertools import imap
from operator import mul

from coord import distfromvect


def add_list_key(d, k, v):
	if not d.has_key(k):
		d[k] = []
	d[k].append(v)

def add_unique(d, k, l):
	if not d.has_key(k):
		d[k] = []
	for v in l:
		if v not in d[k]:
			d[k].append(v)

def display(data):

	d = data.keys()
	d.sort()
	for x, y in d:
		print x, y, data[x,y]

def by_val(kv1, kv2):
	k1, v1 = kv1
	k2, v2 = kv2
	return cmp(v1, v2)

def by_key(kv1, kv2):
	k1, v1 = kv1
	k2, v2 = kv2
	return cmp(k1, k2)

def sorted_index(d):
	l = d.items()
	l.sort(by_val)
	return map(lambda (x,y): x, l)

def sign(n):
	return cmp(n, 0)

def shuffle(i0, reorder):

	print i0
	print reorder

	idx = {}

	id = {}
	ik = []
	for (k, v) in i0:
		ik.append(k)
		id[k] = v

	for k in ik:
		idx[k] = k

	print id
	print idx

	new_order = []
	for i in range(len(idx)):
		new_i = idx[reorder[i]]
		new_order.append((new_i, id[new_i]))

	return new_order

class dataset(dict):

	def __init__(self, max_x=0, max_y=0):

		self.max_x = max_x
		self.max_y = max_y
		self.row_order = []
		self.col_order = []

	def write_file(self, fname, mul=1, repeat=0, avg=0):

		f = open(fname, "w")
		for x in range(self.max_x):
			for y in range(self.max_y):
				if self.has_key((x,y)):
					for m in range(mul):
						for n in range(mul):
							f.write("%f\t%f\t%f\n" % (x+m*self.max_x,
												y+n*self.max_y, self[x, y]))
		if repeat:
			for m in range(-repeat,repeat+mul*self.max_x):
				for n in range(-repeat,repeat+mul*self.max_y):
					if 0 <= m and m <= mul*self.max_x and 0 <= n and n <= mul*self.max_y:
						continue
					f.write("%f\t%f\t%f\n" % (m, n, avg))
			
		f.close()

	def read_file(self, fname):

		self.max_x = 0
		self.max_y = 0
		f = open(fname)
		for l in f.readlines():
			l = l.strip().split('\t')
			l = map(lambda x: x[:-1], l)
			l = map(float, l)
			x = int(l[0])
			y = int(l[1])
			self[x, y] = float(l[2])
			self.max_x = max(x+1, self.max_x)
			self.max_y = max(y+1, self.max_y)
		self.row_order = zip(range(self.max_y), range(self.max_y))
		self.col_order = zip(range(self.max_x), range(self.max_x))

	def select_col(self, col):

		r = {}
		for y in range(self.max_y):
			if self.has_key((col,y)):
				r[y] = self[col, y]
		return r

	def select_row(self, row):

		r = {}
		for x in range(self.max_x):
			if self.has_key((x,row)):
				r[x] = self[x, row]
		return r


	def reorder_cols(self, i0):

		res = dataset(self.max_x, self.max_y)
		res.col_order = shuffle(self.col_order, i0)
		res.row_order = copy(self.row_order)
		idx = {}

		isort = copy(i0)
		isort.sort()

		i0 = copy(i0)
		for i in range(len(i0)):
			idx[isort.pop(0)] = i0.pop(0)

		for y in range(self.max_y):
			for (x, idx_x) in idx.items():
				if self.has_key((idx_x, y)):
					res[x,y] = self[idx[x], y]
		return res

	def reorder_rows(self, i0):

		res = dataset(self.max_x, self.max_y)
		print "reorder_rows", self.row_order
		res.row_order = shuffle(self.row_order, i0)
		res.col_order = copy(self.col_order)
		idx = {}

		isort = copy(i0)
		isort.sort()

		i0 = copy(i0)
		for i in range(len(i0)):
			idx[isort.pop(0)] = i0.pop(0)

		for x in range(self.max_x):
			for (y, idx_y) in idx.items():
				if self.has_key((x, idx_y)):
					res[x,y] = self[x, idx[y]]
		return res

	def signed_std_dev(self):

		avg = self.dot_average()

		newd = dataset(self.max_x, self.max_y)
		newd.row_order = copy(self.row_order)
		newd.col_order = copy(self.col_order)

		for x in range(0, self.max_x):
			for y in range(0, self.max_y):
				if not self.has_key((x, y)):
					continue
				dv = self[x, y] - avg 
				if dv < 0:
					newd[x, y] = -sqrt(dv * dv)
				else:
					newd[x, y] = sqrt(dv * dv)

		return newd

	def std_dev(self):

		avg = self.dot_average()

		newd = dataset(self.max_x, self.max_y)
		newd.row_order = copy(self.row_order)
		newd.col_order = copy(self.col_order)

		for x in range(0, self.max_x):
			for y in range(0, self.max_y):
				if not self.has_key((x, y)):
					continue
				dv = self[x, y] - avg 
				newd[x, y] = sqrt(dv * dv)

		return newd

	def make_z_ratios(self, test_x, test_y, max_height, allow_err, exclude={}):

		test_height = self[test_x, test_y]
		res = {}
		for x in range(0, self.max_x):
			for y in range(0, self.max_y):
				if not self.has_key((x, y)):
					continue
				if y < test_y:
					continue
				if y == test_y and x <= test_x:
					continue
				dist_x = x - test_x
				dist_y = y - test_y

				if exclude.has_key(pi/2):
					ex_x, ex_y = exclude[pi/2]
					if ex_x is not None and x == ex_x and dist_x == 0:
						print "excluding x", degrees(pi/2), x, y, test_x, test_y
						continue
				if exclude.has_key(0.0):
					ex_x, ex_y = exclude[0.0]
					if ex_y is not None and y == ex_y and dist_y == 0:
						print "excluding y", degrees(0.0), x, y, test_x, test_y
						continue

				dist = sqrt(dist_x*dist_x + dist_y*dist_y)
				height_diff = 100 * (self[x,y] - test_height) / max_height

				ang = abs(atan(height_diff / dist))

				ang = float(int(ang/allow_err)*allow_err) # round to err range

				#print test_x, test_y, x, y, self[x,y], degrees(ang)

				add_list_key(res, ang, (test_x, test_y, x, y))

		return res

	def z_ratio_detect(self, allow_err, exclude={}):

		max_height = 0
		for x in range(0, self.max_x):
			for y in range(0, self.max_y):
				if not self.has_key((x, y)):
					continue
				max_height = max(self[x,y], max_height)

		res = {}
		for x in range(0, self.max_x):
			for y in range(0, self.max_y):
				if not self.has_key((x, y)):
					continue

				rats = self.make_z_ratios(x, y, max_height, allow_err, exclude)
				for ang, locs in rats.items():
					for xy in locs:
						add_list_key(res, ang, xy)
		return res

	def find_close_values(self, x1, y1, x2, y2, dist_thresh):
		"""	restricted to 0, 45 and 90 degrees
		"""

		res = []

		v1 = (x1, y1, self[x1, y1])
		v2 = (x2, y2, self[x2, y2])

		for x in range(x1, x2+1):
			for y in range(y1, y2+1):
				if x == x1 and y == y1:
					continue
				if x == x2 and y == y2:
					continue
				#if x != x2 or y != y2 or abs(x-x2) != abs(y-y2):
				#	continue
				#print x1, y1, x2, y2
				if not self.has_key((x,y)):
					continue
				v0 = (x, y, self[x, y])
				if distfromvect(v1, v2, v0) <= dist_thresh:
					res.append((x, y))
		return res

	def find_lines(self, ratios):

		res = {}
		print "num ratios", len(ratios)
		for angles, vectors in ratios.items():
			#print "angles", angles, len(vectors)
			for (x1, y1, x2, y2) in vectors:
				add_unique(res, angles, [(x1, y1)])
				add_unique(res, angles, [(x2, y2)])
				points = self.find_close_values(x1, y1, x2, y2, 0.01)
				if not points:
					continue
				add_unique(res, angles, points)
		return res

	def rotate(self, rotate_x, rotate_y):
		"""	rotate ( >> or << )
		"""

		newd = dataset(self.max_x, self.max_y)

		i0 = range(self.max_x)
		i0 = i0[rotate_x:] + i0[:rotate_x]
		newd.col_order = shuffle(self.col_order, i0)

		i0 = range(self.max_y)
		i0 = i0[rotate_y:] + i0[:rotate_y]
		newd.row_order = shuffle(self.row_order, i0)

		for x in range(self.max_x):
			for y in range(self.max_y):

				new_x = (x + self.max_x + rotate_x) % self.max_x
				new_y = (y + self.max_y + rotate_y) % self.max_y
	
				test_x, test_y = newd.order_map(x, y)
				assert test_x == new_x and test_y == new_y

				newd[new_x, new_y] = self[x, y]

		return newd
		
	def order_map(self, x, y):

		return self.col_order[x][1], self.row_order[y][1]

		for i in range(len(self.col_order)):
			k, v = self.col_order[i]
			if not isinstance(v, list):
				v = [v]
			#print x, k, v
			if x == k:
				new_x = i
				break
		for i in range(len(self.row_order)):
			k, v = self.row_order[i]
			if not isinstance(v, list):
				v = [v]
			#print y, k, v
			if y in v:
				new_y = i
				break
		return new_x, new_y

	def remap(self, bitstep_x, bitstep_y):
		"""	take every bitstep_xth and bitstep_yth value in sequence,
			then bitstep_xth+1 etc.
		"""

		newd = dataset(self.max_x, self.max_y)

		x = 0
		i0 = []
		for step_x in range(bitstep_x):
			for new_x in range(step_x, self.max_x, bitstep_x):
				i0.append(new_x)
		print "remap col", i0
		newd.col_order = shuffle(self.col_order, i0)
		print "remapped col", newd.col_order

		y = 0
		i0 = []
		for step_y in range(bitstep_y):
			for new_y in range(step_y, self.max_y, bitstep_y):
				i0.append(new_y)
		newd.row_order = shuffle(self.row_order, i0)

		x = 0

		for step_x in range(bitstep_x):
			for new_x in range(step_x, self.max_x, bitstep_x):

				y = 0

				for step_y in range(bitstep_y):
					for new_y in range(step_y, self.max_y, bitstep_y):

						test_x, test_y = newd.order_map(x, y)
						#print x, y, test_x, test_y, new_x, new_y
						assert test_x == new_x and test_y == new_y

						#newd[x, y] = self[new_x, new_y]
						newd[new_x, new_y] = self[x, y]

						y += 1
			
				x += 1

		return newd
		
	def sub_average(self, avg):
		"""	
		"""

		newd = dataset(self.max_x, self.max_y)

		for new_x in range(0, self.max_x):
			for new_y in range(0, self.max_y):

				if avg.has_key((new_x, new_y)):
					total = avg[new_x, new_y]

				newd[new_x, new_y] = self[new_x, new_y] - total

		newd.row_order = copy(self.row_order)
		newd.col_order = copy(self.col_order)

		return newd

	def average(self, block_x, block_y):

		newd = dataset(self.max_x, self.max_y)
		blocksize = block_x * block_y

		for new_x in range(0, self.max_x, block_x):
			for new_y in range(0, self.max_y, block_y):

				total = 0.0

				for step_x in range(block_x):
					for step_y in range(block_y):

						total += self[new_x+step_x, new_y+step_y]

				newd[new_x, new_y] = total / blocksize

		newd.row_order = []
		newd.col_order = []

		# averaging assumes a "straight" (0,0), (1,1) .. etc. set
		for new_x in range(0, self.max_x, block_x):
			v = []
			for step_x in range(block_x):
				v.append(self.col_order[new_x+step_x][1]) # drop the index!
			newd.col_order.append((new_x, v))

		for new_y in range(0, self.max_y, block_y):
			v = []
			for step_y in range(block_y):
				v.append(self.row_order[new_y+step_y][1]) # drop the index!
			newd.row_order.append((new_y, v))

		return newd

	def weightaverage(data, max_x, max_y, block_x, block_y):

		newd = dataset(self.max_x, self.max_y)
		blocksize = block_x * block_y

		range_total = 0.0
		for step_x in range(-block_x, block_x+1):
			for step_y in range(-block_y, block_y+1):

				sqdist = (step_x*step_x) + (step_y*step_y)
				rdist = block_x * block_y / (20.0+sqdist)
				range_total += rdist

				#print step_x, step_y, rdist

		print range_total

		for new_x in range(0, self.max_x):
			for new_y in range(0, self.max_y):

				total = 0.0

				for step_x in range(-block_x, block_x+1):
					for step_y in range(-block_y, block_y+1):

						sqdist = (step_x*step_x) + (step_y*step_y)
						dist = block_x * block_y / (1+sqdist)

						nx = (new_x+step_x+max_x) % max_x
						ny = (new_y+step_y+max_y) % max_y
						total += self[nx, ny] * dist / range_total

				newd[new_x, new_y] = total / blocksize

		return newd

	def dot_average(self):

		total = 0.0
		num = 0.0
		for x in range(0, self.max_x):
			for y in range(0, self.max_y):
				if self.has_key((x,y)):
					total += self[x,y]
					num += 1
		return total / num

def remove_col_from_ratios(res, col_num):
	"""	when sorting the histogram (by col) then obviously, that col's
		statistics will be complete rubbish.  we specifically need to
		exclude it from the analysis.

		the algorithm is rather overzealous.
	"""
	angles = res.keys()
	angles.sort()
	for (ang, theta) in angles:
		vectors = res[ang,theta]
		maybe_to_delete = []
		for (x, y) in vectors:
			if col_num == x:
				maybe_to_delete.append((x, y))
		if len(maybe_to_delete) < 2:
			continue
		print "removing", degrees(ang), degrees(theta), maybe_to_delete, vectors
		for xy in maybe_to_delete:
			vectors.remove(xy)

def remove_row_from_ratios(res, row_num):
	"""	when sorting the histogram (by row) then obviously, that row's
		statistics will be complete rubbish.  we specifically need to
		exclude it from the analysis.

		the algorithm is rather overzealous.
	"""
	angles = res.keys()
	angles.sort()
	for (ang, theta) in angles:
		vectors = res[ang,theta]
		maybe_to_delete = []
		for (x, y) in vectors:
			if row_num == y:
				maybe_to_delete.append((x, y))
		if len(maybe_to_delete) < 2:
			continue
		print "removing", degrees(ang), degrees(theta), maybe_to_delete, vectors
		for xy in maybe_to_delete:
			vectors.remove(xy)

def print_ratios(res, list_len=0):

	angles = res.keys()
	angles.sort()
	for (ang, theta) in angles:
		vectors = res[ang,theta]
		if len(vectors) > list_len:
			print "Angle: ", degrees(ang), degrees(theta), vectors

def theta_ratios_detect(data, allow_err):

	res = {}
	for ang, vectors in data.items():
		theta_vectors = make_theta_ratios(vectors, allow_err)
		for theta, vectors in theta_vectors.items():
			for xy in vectors:
				add_list_key(res, (ang, theta), xy)
	return res

def make_theta_ratios(vectors, allow_err):

	res = {}
	print len(vectors)
	for vector in vectors:
		(test_x, test_y, x, y) = vector
		xx = float(x - test_x)
		yy = float(y - test_y)

		if xx == 0:
			ang = sign(yy) * pi / 2
		else:
			ang = atan(yy/xx) - sign(xx)*sign(yy)*pi*(xx<0)

		ang = float(int(ang/allow_err)*allow_err) # round to err range

		if int(degrees(ang)) % 45 != 0:
			continue

		add_list_key(res, ang, vector)

	return res

def test():

	#order = zip(range(4), range(4))
	#reorder = [3,2,0,1]
	#order = [(0, [0, 16, 32, 48, 64, 80, 96, 112]), (96, [12, 28, 44, 60, 76, 92, 108, 124]), (8, [1, 17, 33, 49, 65, 81, 97, 113]), (112, [14, 30, 46, 62, 78, 94, 110, 126])]
	#reorder = [112, 96, 0, 8]

	#print order
	#print shuffle(order, reorder)
	#sys.exit(0)
		
	os.chdir("graphs")

	fname = "test.0.0.2304.txt"
	#fname = "test.1.1.2048.txt"
	bitstep_x = 2
	bitstep_y = 2

	ds = dataset()
	ds.read_file(fname)
	#print d

	#d = remap(d, max_x, max_y, 8, 7)
	#ds = ds.remap(16, 16)
	#ds = ds.rotate(8, 8)
	#d = remap(d, max_x, max_y, 32, 32)
	#d = remap(d, max_x, max_y, 64, 64)
	#d = remap(d, max_x, max_y, 2, 2)
	#d = remap(d, max_x, max_y, 4, 4)

	d = ds.average(16, 16)
	#d = ds.average(8, 7)
	d = d.signed_std_dev()
	#print d.row_order

	r = d.select_col(32)
	i = sorted_index(r)
	print "reorder:", r
	d = d.reorder_rows(i)

	r = d.select_row(16)
	i = sorted_index(r)
	print "reorder:", r
	d = d.reorder_cols(i)
	print d.row_order

	#sys.exit(0)

	#d = remap(d, max_x, max_y, bitstep_x, bitstep_y)
	#d = std_dev(d, max_x, max_y)
	#d = average(d, max_x, max_y, bitstep_x, bitstep_y)
	#d = weightaverage(d, max_x, max_y, 8, 8)
	#display(newd)

	avg = d.dot_average()

	#write_file("remap.%dx%d.%s" % (bitstep_x, bitstep_y, fname), d, max_x, max_y,
	#			mul=3,repeat=64, avg=avg)
	#write_file("remap.%dx%d.%s" % (bitstep_x, bitstep_y, fname), d, max_x, max_y,
	#			repeat=4, avg=avg)
	d.write_file("remap.%dx%d.%s" % (bitstep_x, bitstep_y, fname))
		
	allow_err = pi/2/90
	r = d.z_ratio_detect(allow_err)
	r = theta_ratios_detect(r, allow_err)
	#print_ratios(r, 4)
	r = d.find_lines(r)
	print_ratios(r, 11)

	sys.exit(0)

	print "creating random file"
	from random import Random
	r = Random()
	d = {}
	for x in range(128):
		for y in range(128):
			d[x, y] = 0

	for i in range(256000):
		x = r.randint(0, 127)
		y = r.randint(0, 127)
		rv = r.randint(0, 10)
		d[x, y] += rv
		if i % 1000 == 0:
			print i

	write_file("random.txt", d, 128, 128)

if __name__ == '__main__':
	test()
